@@ -24,9 +24,9 @@
 //! ```
 
 use std::{
-    iter::StepBy,
-    ops::{Index, IndexMut, RangeBounds},
-    slice::{Iter, IterMut},
+    iter::{StepBy, Take},
+    ops::{Bound, Index, IndexMut, RangeBounds},
+    slice::{ChunksMut, Iter, IterMut},
 };
 
 use glam::{IVec2, UVec2};
@@ -90,7 +90,7 @@ impl<T: Clone> Grid<T> {
     ///
     /// Goes from left to right.
     #[inline]
-    pub fn row_iter_mut(&mut self, y: usize) -> IterMut<T> {
+    pub fn row_iter_mut(&mut self, y: usize) -> impl Iterator<Item = &mut T> {
         let w = self.width() as usize;
         let i = y * w;
         self.data[i..i + w].iter_mut()
@@ -99,14 +99,14 @@ impl<T: Clone> Grid<T> {
     /// Insert into a row of the grid using an iterator.
     ///
     /// Will insert up to the length of a row.
-    pub fn insert_row(&mut self, y: usize, row: impl IntoIterator<Item = T>) {
+    pub fn insert_row(&mut self, y: usize, row: impl Iterator<Item = T>) {
         self.insert_row_at([0, y as i32], row);
     }
 
     /// Insert into a row of the grid using an iterator.
     ///
     /// Will insert up to the length of a row.
-    pub fn insert_row_at(&mut self, xy: [i32; 2], row: impl IntoIterator<Item = T>) {
+    pub fn insert_row_at(&mut self, xy: [i32; 2], row: impl Iterator<Item = T>) {
         let [x, y] = xy;
         let iter = self.row_iter_mut(y as usize).skip(x as usize);
         for (v, input) in iter.zip(row) {
@@ -237,6 +237,7 @@ impl<T: Clone> Grid<T> {
         }
     }
 
+    #[inline]
     pub fn is_in_bounds(&self, pos: IVec2) -> bool {
         pos.cmpge(IVec2::ZERO).all() && pos.cmplt(self.size().as_ivec2()).all()
     }
@@ -305,6 +306,65 @@ impl<T: Clone> Grid<T> {
         let (min, max) = ranges_to_min_max_usize(slice, self.len());
         &mut self.data[min..max]
     }
+
+    /// Iterate over a range of rows.
+    ///
+    /// Yields &\[T\] (Slice of T)
+    pub fn rows_iter(&self, range: impl RangeBounds<usize>) -> impl Iterator<Item = &[T]> {
+        let [start, end] = self.range_to_start_end(range, 1);
+        let width = self.width() as usize;
+        let x = start * width;
+        let count = end.saturating_sub(start) + 1;
+        let chunks = self.data[x..].chunks(width);
+        chunks.take(count)
+    }
+
+    /// Iterate mutably over a range of rows.
+    ///
+    /// Yields &mut \[`T`\] (Slice of mutable `T`)
+    pub fn rows_iter_mut(&mut self, range: impl RangeBounds<usize>) -> Take<ChunksMut<T>> {
+        let [start, end] = self.range_to_start_end(range, 1);
+        let width = self.width() as usize;
+        let x = start * width;
+        let count = end - start + 1;
+        let chunks = self.data[x..].chunks_mut(width);
+        chunks.take(count)
+    }
+
+    /// Final index along a given axis, where 0 == width, and 1 == height.
+    pub fn axis_index(&self, axis: usize) -> usize {
+        match axis {
+            0 => self.right_index(),
+            1 => self.top_index(),
+            _ => panic!("Invalid grid axis {}", axis),
+        }
+    }
+
+    pub fn axis_size(&self, axis: usize) -> usize {
+        match axis {
+            0 => self.width() as usize,
+            1 => self.height() as usize,
+            _ => panic!("Invalid grid axis {}", axis),
+        }
+    }
+
+    /// Convert a range into a [start,end] pair.
+    ///
+    /// An unbounded "end" returned by this function should be treated as EXCLUSIVE.
+    fn range_to_start_end(&self, range: impl RangeBounds<usize>, axis: usize) -> [usize; 2] {
+        let start = match range.start_bound() {
+            Bound::Included(start) => *start,
+            Bound::Excluded(start) => *start,
+            Bound::Unbounded => 0,
+        };
+        let end = match range.end_bound() {
+            Bound::Included(end) => *end,
+            Bound::Excluded(end) => *end - 1,
+            Bound::Unbounded => self.axis_size(axis),
+        };
+
+        [start, end]
+    }
 }
 
 fn ranges_to_min_max_usize<RANGE: RangeBounds<usize>>(range: RANGE, max: usize) -> (usize, usize) {
@@ -321,7 +381,7 @@ fn ranges_to_min_max_usize<RANGE: RangeBounds<usize>>(range: RANGE, max: usize)
     };
 
     debug_assert!(range_min < range_max);
-    debug_assert!(range_max < max);
+    debug_assert!(range_max <= max);
 
     (range_min, range_max)
 }
@@ -415,6 +475,39 @@ impl<T: Clone> IndexMut<UVec2> for Grid<T> {
 mod tests {
     use super::*;
 
+    #[test]
+    fn range_convert() {
+        let grid = Grid::new(0, [5, 11]);
+        let [start, end] = grid.range_to_start_end(.., 0);
+        assert_eq!([start, end], [0, 5]);
+        let [start, count] = grid.range_to_start_end(5..=10, 0);
+        assert_eq!([start, count], [5, 10]);
+        let [start, count] = grid.range_to_start_end(3..11, 0);
+        assert_eq!([start, count], [3, 10]);
+    }
+
+    #[test]
+    fn rows_iter() {
+        let mut grid = Grid::default([3, 10]);
+        grid.insert_row(3, std::iter::repeat(5));
+        grid.insert_row(4, std::iter::repeat(6));
+
+        let mut iter = grid.rows_iter(3..=4);
+
+        assert!(iter.next().unwrap().iter().all(|v| *v == 5));
+        assert!(iter.next().unwrap().iter().all(|v| *v == 6));
+    }
+
+    #[test]
+    fn rows_iter_mut() {
+        let mut grid = Grid::default([3, 4]);
+        for row in grid.rows_iter_mut(..) {
+            row.iter_mut().for_each(|v| *v = 5);
+        }
+
+        assert!(grid.iter().all(|v| *v == 5));
+    }
+
     #[test]
     fn row_iter() {
         let mut grid = Grid::default([10, 15]);
@@ -471,7 +564,7 @@ mod tests {
         assert_eq!(1, p.x);
         assert_eq!(0, p.y);
 
-        let (p, _) = iter.skip(3).next().unwrap();
+        let (p, _) = iter.nth(3).unwrap();
         assert_eq!(0, p.x);
         assert_eq!(1, p.y);
     }
@@ -535,7 +628,7 @@ mod tests {
 
         let (p, _) = iter.next().unwrap();
         assert_eq!(p, IVec2::new(2, 2));
-        assert_eq!(iter.skip(7).next().unwrap().0, IVec2::new(4, 4));
+        assert_eq!(iter.nth(7).unwrap().0, IVec2::new(4, 4));
     }
 
     #[test]
